#! /usr/bin/env python
import random
import re
from time import sleep


from simdtk import barf


try:
	from gevent import socket
except ImportError:
	import socket


class IRCConnection(object):
	"""
	Connection class for connecting to IRC servers
	"""
	# a couple handy regexes for reading text
	nick_re = re.compile(
		'.*?Nickname is already in use'
	)
	nick_change_re = re.compile(
		':(?P<old_nick>.*?)!\S+\s+?NICK\s+:\s*(?P<new_nick>[-\w]+)'
	)
	ping_re = re.compile('^PING (?P<payload>.*)')
	chanmsg_re = re.compile(
		':(?P<nick>.*?)!\S+\s+?PRIVMSG\s+(?P<channel>#+[-\w]+)\s+:' +
		'(?P<message>[^\n\r]+)'
	)
	privmsg_re = re.compile(
		':(?P<nick>.*?)!~\S+\s+?PRIVMSG\s+[^#][^:]+:(?P<message>[^\n\r]+)'
	)
	part_re = re.compile(
		':(?P<nick>.*?)!\S+\s+?PART\s+(?P<channel>#+[-\w]+)'
	)
	join_re = re.compile(
		':(?P<nick>.*?)!\S+\s+?JOIN\s+.*?(?P<channel>#+[-\w]+)'
	)
	quit_re = re.compile(
		':(?P<nick>.*?)!\S+\s+?QUIT\s+.*'
	)
	registered_re = re.compile(
		':(?P<server>.*?)\s+(?:376|422)'
	)

	def __init__(self, server, port, nick, logfile=None, needs_registration=True):
		self.server = server
		self.port = port
		self.nick = self.base_nick = nick

		self.logfile = logfile
		self._registered = not needs_registration
		self._out_buffer = []
		self._callbacks = []

	def send(self, data, force=False):
		"""
		Send raw data over the wire if connection is registered. Otherewise,
		save the data to an output buffer for transmission later on.
		If the force flag is true, always send data, regardless of
		registration status.
		"""
		if self._registered or force:
			self._sock_file.write('%s\r\n' % data)
			self._sock_file.flush()
		else:
			self._out_buffer.append(data)

	def connect(self):
		"""
		Connect to the IRC server using the nickname
		"""
		self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

		# Try some SSL socket wrapping.
		# This doesn't enforce certificate requirements. :(
		try:
			import ssl
			self._sock = ssl.wrap_socket(self._sock)
		except Exception as e:
			print(e)

		try:
			self._sock.connect((self.server, self.port))
		except socket.error:
			barf.error(
				'Unable to connect to %s on port %d' % (
					self.server,
					self.port
				),
			)

			return False

		self._sock_file = self._sock.makefile('rw')
		self.register_nick()
		self.register()
		return True

	def close(self):
		self._sock.close()

	def register_nick(self):
		barf.action('Registering nick %s' % self.nick)
		self.send('NICK %s' % self.nick, True)

	def register(self):
		barf.action('Authing as %s' % self.nick)
		self.send('USER %s %s bla :%s' % (self.nick, self.server, self.nick), True)

	def join(self, channel):
		if not channel.startswith('#'):
			channel = '#%s' % channel
		self.send('JOIN %s' % channel)
		barf.action('joining %s' % channel)

	def part(self, channel):
		if not channel.startswith('#'):
			channel = '#%s' % channel
		self.send('PART %s' % channel)
		barf.action('leaving %s' % channel)

	def quit(self, msg):
		self.send('QUIT')
		self.respond(msg)

	def respond(self, message, channel=None, nick=None):
		"""
		Multipurpose method for sending responses to channel or via message to
		a single user
		"""
		try:
			if channel:
				if not channel.startswith('#'):
					channel = '#%s' % channel
				barf.message('%s: <%s> \033[0m%s' % (channel, self.nick, message))
				self.send('PRIVMSG %s :%s' % (channel, message))
			elif nick:
				barf.message('%s: <%s> \033[0m%s' % (nick, self.nick, message))
				self.send('PRIVMSG %s :%s' % (nick, message))
		except UnicodeError as e:
			barf.error('IRCConnection.respond: %s' % e)
			barf.action('Skipping response')
			pass

	def dispatch_patterns(self):
		"""
		Low-level dispatching of socket data based on regex matching, in general
		handles

		* In event a nickname is taken, registers under a different one
		* Responds to periodic PING messages from server
		* Dispatches to registered callbacks when
			- any user leaves or enters a room currently connected to
			- a channel message is observed
			- a private message is received
		"""
		return (
			(self.nick_re, self.new_nick),
			(self.nick_change_re, self.handle_nick_change),
			(self.ping_re, self.handle_ping),
			(self.part_re, self.handle_part),
			(self.join_re, self.handle_join),
			(self.quit_re, self.handle_quit),
			(self.chanmsg_re, self.handle_channel_message),
			(self.privmsg_re, self.handle_private_message),
			(self.registered_re, self.handle_registered),
		)

	def register_callbacks(self, callbacks):
		"""
		Hook for registering custom callbacks for dispatch patterns
		"""
		self._callbacks.extend(callbacks)

	def new_nick(self):
		"""
		Generates a new nickname based on original nickname followed by a
		random number
		"""
		old = self.nick
		self.nick = '%s_%s' % (self.base_nick, random.randint(1, 1000))
		barf.error('Nick %s already taken, trying %s' % (old, self.nick))
		self.register_nick()
		self.handle_nick_change(old, self.nick)

	def handle_nick_change(self, old_nick, new_nick):
		for pattern, callback in self._callbacks:
			if pattern.match('/nick'):
				callback(old_nick, '/nick', new_nick)

	def handle_ping(self, payload):
		"""
		Respond to periodic PING messages from server
		"""
		# We should probably make debug for this file...
		# barf.message('server ping: %s' % payload)
		self.send('PONG %s' % payload, True)

	def handle_registered(self, server):
		"""
		When the connection to the server is registered, send all pending
		data.
		"""
		if not self._registered:
			barf.message('Connected to %s' % server)
			self._registered = True
			for data in self._out_buffer:
				self.send(data)
			self._out_buffer = []

	def handle_part(self, nick, channel):
		for pattern, callback in self._callbacks:
			if pattern.match('/part'):
				callback(nick, '/part %s' % (channel), channel)

	def handle_join(self, nick, channel):
		for pattern, callback in self._callbacks:
			if pattern.match('/join'):
				callback(nick, '/join %s' % (channel), channel)

	def handle_quit(self, nick):
		for pattern, callback in self._callbacks:
			if pattern.match('/quit'):
				callback(nick, '%s /quit' % nick, None)

	def _process_command(self, nick, message, channel):
		results = []

		for pattern, callback in self._callbacks:
			match = pattern.match(message) or pattern.match('/privmsg')
			if match:
				results.append(callback(nick, message, channel, **match.groupdict()))

		return results

	def handle_channel_message(self, nick, channel, message):
		for result in self._process_command(nick, message, channel):
			if result:
				self.respond(result, channel=channel)

	def handle_private_message(self, nick, message):
		for result in self._process_command(nick, message, None):
			if result:
				self.respond(result, nick=nick)

	def enter_event_loop(self):
		"""
		Main loop of the IRCConnection - reads from the socket and dispatches
		based on regex matching
		"""
		patterns = self.dispatch_patterns()

		while 1:
			try:
				data = self._sock_file.readline()
			except socket.error:
				data = None
			if not data:
				barf.action('server closed connection')
				self.close()
				sleep(5)
				return True

			data = data.rstrip()

			for pattern, callback in patterns:
				match = pattern.match(data)
				if match:
					callback(**match.groupdict())


class IRCBot(object):
	"""
	A class that interacts with the IRCConnection class to provide a simple way
	of registering callbacks and scripting IRC interactions
	"""
	def __init__(self, conn):
		self.conn = conn

		# register callbacks with the connection
		self.register_callbacks()

	def register_callbacks(self):
		"""
		Hook for registering callbacks with connection -- handled by __init__()
		"""
		self.conn.register_callbacks((
			(re.compile(pattern), callback)
			for pattern, callback in self.command_patterns()
		))

	def _ping_decorator(self, func):
		def inner(nick, message, channel, **kwargs):
			message = re.sub('^%s[:,\s]\s*' % self.conn.nick, '', message)
			return func(nick, message, channel, **kwargs)
		return inner

	def is_ping(self, message):
		return re.match('^%s[:,\s]' % self.conn.nick, message) is not None

	def fix_ping(self, message):
		return re.sub('^%s[:,\s]\s*' % self.conn.nick, '', message)

	def ping(self, pattern, callback):
		return (
			'^%s[:,\s]\s*%s' % (self.conn.nick, pattern.lstrip('^')),
			self._ping_decorator(callback),
		)

	def command_patterns(self):
		"""
		Hook for defining callbacks, stored as a tuple of 2-tuples:
		return (
			('/join', self.room_greeter),
			('/part', self.room_leaver),
			('!find (^\s+)', self.handle_find),
		)
		"""
		raise NotImplementedError

	def respond(self, message, channel=None, nick=None):
		"""
		Wraps the connection object's respond() method
		"""
		if message is not None:
			self.conn.respond(message, channel, nick)


def run_bot(bot_class, host, port, nick, channels=None):
	"""
	Convenience function to start a bot on the given network, optionally joining
	some channels
	"""
	conn = IRCConnection(host, port, nick)
	bot_instance = bot_class(conn)  # NOQA

	while 1:
		if not conn.connect():
			break

		channels = channels or []

		for channel in channels:
			conn.join(channel)

		conn.enter_event_loop()


class SimpleSerialize(object):
	"""
	Allow simple serialization of data in IRC messages with minimum of space.

	* Only supports dictionaries *
	"""
	def serialize(self, dictionary):
		return '|'.join(('%s:%s' % (k, v) for k, v in dictionary.iteritems()))

	def deserialize(self, string):
		return dict((piece.split(':', 1) for piece in string.split('|')))
